#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Sep 18 16:38:48 2017

@author: db242421


Le module data_handling fournit des fonctions commode pour stocker les differentes informations relatives au fichiers
utiles (irmf, atlas individuels...) dans une structure facilement utilisable de dictionnaire.

"""

